#define ADDR__BASE  10          //   EEPROM      
#define ADDR__TIMER_START         (ADDR__BASE+0) //  EEPROM
#define ADDR__TIMER_TOFLAG_COUNT  (ADDR__BASE+1)
#define ADDR__PULSE_LENGTH        (ADDR__BASE+2)
#define ADDR__RECALC_LEN_0        (ADDR__BASE+3) //  
#define ADDR__RECALC_LEN_1        (ADDR__BASE+4)

#define PIN_OUT      13
#define OSCILL_POINTS 40          //    
#define OSCILL_POINTS_PAUSE 1     //    OSCILL_POINTS_PAUSE * OSCILL_POINTS < 256!!!
#define MAX_TIMER_START 250    //   

#define KEY_RIGHT   0             // LCD
#define KEY_UP      1
#define KEY_DOWN    2
#define KEY_LEFT    3
#define KEY_SELECT  4
#define NUM_KEYS    5

#define CMD__RECALC_LEN         0
#define CMD__PULSE_LENGTH       1
#define CMD__TIMER_START     2
#define CMD__TIMER_TOFLAG_COUNT 3
#define CMD__STATUS             4
#define COMMANDS_COUNT          (CMD__STATUS+1)

#define FASTADC 1 // defines for setting and clearing register bits
#ifndef cbi       //  FASTADC 1 analogRead   20,     100
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#include <EEPROM.h>
#include <LCD4Bit_mod.h> 
//  ,   ,   ,         
char commandNames[COMMANDS_COUNT][13]={
"RECALC COUNT",
"PULSE LENGTH",
"TIMER START",
"TOFLAG COUNT", 
"STATUS"};
char currCommand=2;
int adc_key_val[5] ={30, 150, 360, 535, 760 };
int adc_key_in, key=-1, oldkey=-1;
bool LCD_WRITING=false, statusPart=false;


unsigned int recalcCount=0,    //    OSCILL_POINTS 
             RECALC_LEN=100; // < 65536    recalcCount    
volatile boolean TIMER_FLAG         = false; //    . ,     
volatile uint8_t TIMER_COUNTER      = 0;   //    
volatile uint8_t TIMER_START     = 10;  //    .    TIMER_START  255,  255+1=0  
volatile uint8_t TIMER_TOFLAG_COUNT = 20;  // <=255          
//       
//    (256-TIMER_START)*( 16MHz/8 = 2Mhz or 0.5us) us
//   TIMER_COUNTER   1
//  TIMER_COUNTER >= TIMER_FLAG_TIME,  

volatile uint8_t *REG_8_12_13;//   
volatile uint8_t MASK_12;     //  12 
volatile uint8_t MASK_13;     //  13 

uint8_t PULSE_LENGTH  = 200; //    

unsigned long oscillMeasureTime=0;
unsigned long akbVoltage=0;  //          
unsigned long oscillPointsAcc[ OSCILL_POINTS ]; //    ""
uint8_t analogReadNum=0;     //   

String sep=":";
String outStr;
String cmd="";

LCD4Bit_mod lcd = LCD4Bit_mod(2); 


ISR( TIMER2_OVF_vect ) { //  
    TIMER_COUNTER++;
    if( TIMER_COUNTER > TIMER_TOFLAG_COUNT ){
        TIMER_FLAG   =true;
        TIMER_COUNTER=0;
    }
    TCNT2=TIMER_START; 
}

void timer2Start( unsigned char tValue ){ //    
    //Timer2 Settings: Timer Prescaler /8, mode 0
    //Timmer clock = 16MHz/8 = 2Mhz or 0.5us
    //The /8 prescale gives us a good range to work with so we just hard code this for now.
    //load the timer for its first cycle
    TCNT2=tValue; 
    TCCR2A = 0;
    TCCR2B = 0<<CS22 | 1<<CS21 | 0<<CS20; 
    //Timer2 Overflow Interrupt Enable   
    TIMSK2 = 1<<TOIE2;
    //load the timer for its first cycle
    TCNT2=tValue; 
}



void serialStrVal( String s, int i ){
    Serial.print( s );
    Serial.println( i );
}
void serialStrVal( String s, unsigned int i ){
    Serial.print( s );
    Serial.println( i );
}
void serialStrVal( String s, float i ){
    Serial.print( s );
    Serial.println( i );
}
void serialStrVal( String s, char i ){
    Serial.print( s );
    Serial.println( i );
}


void do_RECALC_LEN( unsigned int val ){ //   
    RECALC_LEN = min( 0xffff, max( 10, val ));
    EEPROM.write ( ADDR__RECALC_LEN_0, (RECALC_LEN >> 8));
    EEPROM.write ( ADDR__RECALC_LEN_1, (RECALC_LEN&0xFF));
    serialStrVal("m:New RECALC_LEN set to ", RECALC_LEN );
}

void do_TIMER_START( unsigned char val ){ //   
   TIMER_START = min( MAX_TIMER_START, max( 1, val ) );
   EEPROM.write ( ADDR__TIMER_START, TIMER_START);
   serialStrVal("m:TIMER_START set to ", TIMER_START );
}
void do_TIMER_TOFLAG_COUNT( unsigned char val ){ //    
   TIMER_TOFLAG_COUNT = max( 1, val );
   EEPROM.write ( ADDR__TIMER_TOFLAG_COUNT, TIMER_TOFLAG_COUNT );
   serialStrVal("m:TIMER_TOFLAG_COUNT  set to ", TIMER_TOFLAG_COUNT  );
}

void do_PULSE_LENGTH( int val ){ //   
   PULSE_LENGTH = min( 255, max( 0, val ));
   EEPROM.write ( ADDR__PULSE_LENGTH, PULSE_LENGTH );
   serialStrVal("m:PULSE_LENGTH set to ", PULSE_LENGTH );
}

void do_STATUS(){ //  
   Serial.println("");
   serialStrVal("\n\nm:(p=) PULSE_LENGTH=", PULSE_LENGTH );
   serialStrVal("m:(t=) TIMER_START=", TIMER_START );
   serialStrVal("m:(f=) TIMER_FLAG_COUNT=", TIMER_TOFLAG_COUNT );
   serialStrVal("m:(r=) RECALC_LEN=", RECALC_LEN );
   Serial.println("m:(s=) print status" );
   Serial.println("");
}


void doCmd(){
   if( cmd.length()>0 ){
      if( cmd.length()>=140 ){
	 Serial.println( "\nm:Line too long!" );
      }
      cmd.toLowerCase();
      char mnemo =cmd.charAt( 0 );
      int possInt=cmd.substring( 2 ).toInt();
      Serial.print( "c:");
      Serial.println( cmd );
      switch( mnemo ){
        case 'r': do_RECALC_LEN( possInt ); break;
        case 't': do_TIMER_START( possInt ); break;
        case 'p': do_PULSE_LENGTH( possInt ); break;
        case 'f': do_TIMER_TOFLAG_COUNT( possInt ); break;
        case 's': do_STATUS(); break;
      default:
        Serial.println( "Command not found" );
        do_STATUS();
      }
      cmd="";
      lcdStatus( true );
   }
}

void readSerial(){
   while( Serial.available() ){
      char ch=Serial.read();
      if( ch == 10 || ch == 13 || cmd.length()>=140 ){
	 doCmd();
      } else {
	 cmd+=ch ;
      }
   }
}

void loadEEPROMvalues(){
    TIMER_START     = EEPROM.read ( ADDR__TIMER_START);
    TIMER_TOFLAG_COUNT = EEPROM.read ( ADDR__TIMER_TOFLAG_COUNT );
    PULSE_LENGTH       = EEPROM.read ( ADDR__PULSE_LENGTH );    
    RECALC_LEN	       = EEPROM.read ( ADDR__RECALC_LEN_0);
    RECALC_LEN	<<= 8;
    RECALC_LEN	      += EEPROM.read ( ADDR__RECALC_LEN_1);
    //      ,     0xFF
    if( (TIMER_START & TIMER_TOFLAG_COUNT & PULSE_LENGTH & RECALC_LEN ) == 0xFF ) saveEEPROMdefaults();
        
}
void saveEEPROMdefaults(){
    TIMER_START = 10;
    EEPROM.write ( ADDR__TIMER_START, TIMER_START );
    TIMER_TOFLAG_COUNT = 10;
    EEPROM.write ( ADDR__TIMER_TOFLAG_COUNT, TIMER_TOFLAG_COUNT );
    PULSE_LENGTH   = 100;
    EEPROM.write ( ADDR__PULSE_LENGTH, PULSE_LENGTH );    
    RECALC_LEN	  = 100;
    EEPROM.write ( ADDR__RECALC_LEN_0, (RECALC_LEN>>8) );
    EEPROM.write ( ADDR__RECALC_LEN_1, (RECALC_LEN&0xFF));
}


void lcdStatus( boolean dontChangePart ){
    if( !dontChangePart ) statusPart = !statusPart;
    String s="";
    if( statusPart  ){
        s+="p"+String(PULSE_LENGTH);
        s+=" t"+String(TIMER_START);
        s+=" f"+String(TIMER_TOFLAG_COUNT);
    }else{
        s+=" r"+String(RECALC_LEN);
    }
    
    while( s.length()<16 ) s+=" ";
    char c[16];
    s.toCharArray(c, 16);
    
    LCD_WRITING=true;
    lcd.cursorTo(2,0);
    lcd.printIn(c);
    LCD_WRITING=false;
}
void lcdPrintCmd(){
    String s=commandNames[ (int)currCommand ]; //  s    
    int cmdVal=0;
    if(CMD__STATUS != currCommand){ //   status  
        switch( currCommand ){
            case CMD__RECALC_LEN        : cmdVal=RECALC_LEN; break;
            case CMD__TIMER_START       : cmdVal=TIMER_START; break;
            case CMD__TIMER_TOFLAG_COUNT: cmdVal=TIMER_TOFLAG_COUNT; break;
            case CMD__PULSE_LENGTH      : cmdVal=PULSE_LENGTH; break;
            case CMD__STATUS     : ;
        }
        s+=(cmdVal<100?" ":"")+String(cmdVal); //      
    }
    while( s.length()<16 ) s+=" "; //  
    char c[16];
    s.toCharArray( c, 16 );
    LCD_WRITING=true;
    lcd.cursorTo(1,0);
    lcd.printIn( c ); 
    LCD_WRITING=false;
}

void cmdAdd(int inc){
    currCommand+=inc;
    if(currCommand>=COMMANDS_COUNT){
        currCommand=0;
    }
    if(currCommand<0){
        currCommand=COMMANDS_COUNT-1;
    }
    lcdPrintCmd();
}

void cmdValueAdd( int inc ){
    switch( currCommand ){
        case CMD__RECALC_LEN           : do_RECALC_LEN( RECALC_LEN + inc*10); break;
        case CMD__TIMER_START       : do_TIMER_START( TIMER_START + inc ); break;
        case CMD__PULSE_LENGTH         : do_PULSE_LENGTH( PULSE_LENGTH + inc ); break;
        case CMD__TIMER_TOFLAG_COUNT   : do_TIMER_TOFLAG_COUNT( TIMER_TOFLAG_COUNT + inc ); break;
        case CMD__STATUS               : do_STATUS(); break;
    }
    lcdPrintCmd();
    lcdStatus( true );
}

int  get_key( int input){ // Convert ADC value to key number
	int k;
	for (k = 0; k < NUM_KEYS; k++){
		if (input < adc_key_val[k]) return k;
	}
    if (k >= NUM_KEYS) k = -1;     // No valid key pressed
    return k;
}
void doKeys(){
    adc_key_in = analogRead(0);        // read the value from the sensor  
    key = get_key(adc_key_in);		   // convert into key press

    if (key != oldkey){				   // if keypress is detected
        delay(50);		                   // wait for debounce time
        adc_key_in = analogRead(0);    // read the value from the sensor  
        key = get_key(adc_key_in);		   // convert into key press
        if (key != oldkey){
            oldkey = key;
            switch( key ){
                case KEY_RIGHT: cmdValueAdd( 1); break;
                case KEY_UP   : cmdAdd(1); break;
                case KEY_DOWN : cmdAdd(-1); break;
                case KEY_LEFT : cmdValueAdd(-1); break;
                case KEY_SELECT  : lcdStatus( false ); return;
                default:
                    return;
            }
            lcdStatus( true );
        }
    }
}



//  ,       
void myAnalogReadStart(uint8_t pin){
	if (pin >= 14) pin -= 14; // allow for channel or pin numbers
    volatile uint8_t _ADMUX;
    _ADMUX=(DEFAULT << 6) | (pin & 0x07);
	ADMUX = (DEFAULT << 6) | (pin & 0x07);
	sbi(ADCSRA, ADSC);
}

//         
int myAnalogReadGet(){
	uint8_t low, high;
	// ADSC is cleared when the conversion finishes
	while (bit_is_set(ADCSRA, ADSC));
	low  = ADCL;
	high = ADCH;
	return (high << 8) | low; // combine the two bytes
} 



void setup(){
    loadEEPROMvalues();
    for( int i=0; i < OSCILL_POINTS; i++ ) oscillPointsAcc[i]=0;
    REG_8_12_13	= portOutputRegister( digitalPinToPort( 8 ) );
    MASK_12 	= digitalPinToBitMask( 12 );
    MASK_13 	= digitalPinToBitMask( 13 );

    Serial.begin( 57600 );
    Serial.println( "" );
    Serial.println( "#graph,0,1,p,16,AKB,0,0,255" );
    Serial.print( "#oscill,"); Serial.print( OSCILL_POINTS);  Serial.println(",16,AKB Pulse,255,0,0" );
    Serial.println( "#SysMessagesDone" );
    Serial.println( "m:Starting" );

    #if FASTADC // set prescale to 16
      sbi(ADCSRA,ADPS2) ;
      cbi(ADCSRA,ADPS1) ;
      cbi(ADCSRA,ADPS0) ;
    #endif

    LCD_WRITING=true;
        lcd.init();
        lcd.clear();
        cmdAdd( 0 );
        lcdStatus( true );
    LCD_WRITING=false;
   
    pinMode( PIN_OUT,   OUTPUT  );
    timer2Start( TIMER_START );
    do_STATUS();
    
}
volatile uint8_t myR18;


void loop() {
    while( !TIMER_FLAG || LCD_WRITING ){}; //   
        
    {   //        
        //    OSCILL_POINTS     OSCILL_POINTS_PAUSE * (     )
        //          2  = 250nS
        // ..   OSCILL_POINTS_PAUSE=4,      1
        akbVoltage+= analogRead(A1);      //      ,    20
        myR18= analogReadNum*OSCILL_POINTS_PAUSE;

        ADMUX = 0x41; // A1
        asm("ldi r24,lo8(65)");
        asm("sts 124,r24");
        *REG_8_12_13 |= MASK_13;     //      delayMicroseconds,     ))
        asm("lds r24,PULSE_LENGTH");
        asm("ldi r25,lo8(0)");
        asm("call delayMicroseconds");
        if( analogReadNum == 0 ){    //  ,        
                                     //            
            sbi(ADCSRA, ADSC);       //  == myAnalogReadStart(A1),    
           *REG_8_12_13 &= ~MASK_13; //  
            oscillPointsAcc[ 0 ]+=myAnalogReadGet(); //     
        }else{
           *REG_8_12_13 &= ~MASK_13;
          asm("lds r18,myR18");     // 2  for( i=R18; i>0; i-- ) //       analogReadNum
          asm("tst r18");           // 1 
          asm("breq .my_R18_zero"); // 1  , 2  OSCILL_POINTS_PAUSE==0
                                    // ,      5+1nop+2  sbi = 8*0.0625 = 0.5
                                    //    OSCILL_POINTS_PAUSE    0,5
                                    //    ,    
      asm(".my_R18_nonzero:");      // 1   NOP 3 =62.5, => 1 OSCILL_POINTS_PAUSE = 0,5
  
          asm("nop\n nop\n nop\n nop\n nop\n ");    // 5*62,5ns + 3*62.5    0.5
          
          asm("dec r18"); 	                        // 1 
          asm("brne .my_R18_nonzero");              // 2    , 1  
      asm(".my_R18_zero:");
          asm("nop");                               //  ,   
  
          sbi(ADCSRA, ADSC); //        myAnalogReadStart(A1);
          oscillPointsAcc[ analogReadNum ]+=myAnalogReadGet();
        }

        analogReadNum++;
        if( analogReadNum >= OSCILL_POINTS ){ //       ""?
            analogReadNum=0; //    0
            recalcCount++;     //   
        }
        
        if( recalcCount >= RECALC_LEN  ){         
            //   
            //   * OSCILL_POINTS_PAUSE*analogReadNum 

            outStr = String("o:")+ //   
                +((OSCILL_POINTS_PAUSE*OSCILL_POINTS)>>1);
                
            for( int i=0; i<OSCILL_POINTS; i++ ){ 
                outStr +=  sep+String( int( oscillPointsAcc[i] / recalcCount ));
                oscillPointsAcc[i]=0;
            }
            Serial.println( outStr ); //   
            
            outStr = String("g:")+ int( akbVoltage/OSCILL_POINTS/recalcCount );
            akbVoltage=0;
            Serial.println( outStr ); //   
            recalcCount=0;
        }
    }
    readSerial();
    doKeys();
    TIMER_FLAG=false;
}
/*  
Set the ADLAR bit in ADMUX register and read just ADCH
For example,
    uint8_t readADC8(uint8_t channel)
    {
        ADMUX = (channel & 0x1F) | _BV(ADLAR);
        ADCSRA |= _BV(ADSC); // in atmega16 RMW operation actually _clears_ ADIF if it was set before
        while( !(ADCSRA & _BV(ADIF)) ) ;
        return ADCH;
    }
*/

